@@ -1,30 +1,28 @@
 import pandas as pd
-import json, re
+import json, re, asyncio, sys
 from gql import gql
 from gql import Client as gql_Client
-from gql.transport.requests import RequestsHTTPTransport
+from gql.transport.aiohttp import AIOHTTPTransport
 from .types import __VALUE_TYPES__, DATASET, INNER_DATASET, __PATTERNS__, WM__PATTERNS__
-from requests.exceptions import *  
+from gql.transport.exceptions import *
 from datetime import datetime, timedelta
 from time import sleep
+if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith('win'):
+    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
+import nest_asyncio
+nest_asyncio.apply()
 
 class safeGraphError(Exception):
     pass
 
 class HTTP_Client:
     def __init__(self, apikey, max_tries=3):
+        self.ct = 0
         self.df = pd.DataFrame()
         self.lst = [] 
         self.url = 'https://api.safegraph.com/v1/graphql'
         self.apikey = apikey
         self.headers = {'Content-Type': 'application/json', 'apikey': apikey}
-        self.transport = RequestsHTTPTransport(
-            url=self.url, 
-            verify=True, 
-            retries=max_tries,
-            headers=self.headers,
-        )
-        self.client = gql_Client(transport=self.transport, fetch_schema_from_transport=True)
         self.return_type = "pandas"
         self.max_results = 20
         self.errors = []
@@ -74,6 +72,14 @@ class HTTP_Client:
         else:
             raise safeGraphError(f'''*** Unrecognized DateTime! {value}\n must be either one list/str/dict''')
 
+    async def main_runner(self, query):
+        self.ct+=1
+        transport = AIOHTTPTransport(url=self.url, headers=self.headers)
+        async with gql_Client(
+            transport=transport, fetch_schema_from_transport=True,
+        ) as session:
+            return await session.execute(query)
+
     def __change_value_types_pandas(self):
         for key, val in __VALUE_TYPES__.items():
             try:
@@ -93,8 +99,6 @@ class HTTP_Client:
                     lst[key] = val(lst[key])
                 except KeyError:
                     pass
-                # except TypeError:
-                #     import pdb;pdb.set_trace()
 
     def __column_check_raise(self, columns):
         dict_ = {el:0 for el in columns}
@@ -271,7 +275,10 @@ class HTTP_Client:
 
     def __chunks(self):
         """Yield successive n-sized chunks from self.max_results."""
-        lst = [i for i in range(self.max_results)]
+        try:
+            lst = [i for i in range(self.max_results)]
+        except TypeError:
+            raise safeGraphError("|max_results argument|*** TypeError: 'str' object cannot be interpreted as an integer ")
         n = 500
         for i in range(0, len(lst), n):
             yield lst[i:i + n]
@@ -286,38 +293,7 @@ class HTTP_Client:
         __pick__ = False # picks the next value without additional operation; for first check
         pre = 0
         str_ = ""
-        __split__ = re.sub(r'\s+', " ", __query__).split(" ")
-        for s in __split__:
-            pass
-            # v2
-            # if s == "{":
-            #     pre+=1
-            #     str_ += s + "\n" + pre*"\t"
-            # elif s == "}":
-            #     pre-=1
-            #     str_ += s + "\n" + pre*"\t"
-            # else:
-            #     str_+=s
-
-            # v1
-            # if __pick__:
-            #     if s == "{":
-            #         pre+=1
-            #     elif s == "}":
-            #         pre-=1
-            #     str_+= " " + s + "\n" + pre*"\t"
-            #     __pick__ = False
-            # elif s == "{":
-            #     pre+=1
-            #     str_+=" " + s# + "\n" + pre*"\t"
-            # elif s == "}":
-            #     pre-=1
-            #     str_+=" " + s# + "\n" + pre*"\t"
-            # elif s.endswith(":"):
-            #     str_+=" " + s
-            #     __pick__ = True
-            # else:
-            #     str_+= "\n" + pre*"\t" + s
+        __split__ = re.sub(r'\s+', " ", __query__).split(" ");
         print(re.sub(r'\n+', '\n', __query__))
 
     def save(self, path="__default__", return_type="__default__"):
@@ -391,7 +367,22 @@ class HTTP_Client:
         else:
             return df.to_dict("records")
 
-    def lookup(self, product:str, placekeys:list, columns, date='__default__', preview_query:str=False, return_type:str="pandas"):
+    def lookup(self, product:str, placekeys:list, columns, date='__default__', preview_query:bool=False, return_type:str="pandas"):
+        loop = asyncio.get_event_loop()
+        # asyncio.set_event_loop(loop)
+        if type(placekeys) == str:
+            raise "placekeys cannot be string: list required"
+        results = loop.run_until_complete(self.lookup_(product, placekeys, columns, date, preview_query, return_type, cache=True, results=[]))
+        if preview_query:
+            return
+        arr = [self.main_runner(query) for query in results]
+        results = asyncio.gather(*arr, return_exceptions=True)
+        report = loop.run_until_complete(results)
+        ret = loop.run_until_complete(self.lookup_(product, placekeys, columns, date, preview_query, return_type, cache=False, results=report))
+        # loop.close()
+        return ret
+
+    async def lookup_(self, product:str, placekeys:list, columns, date='__default__', preview_query:bool=False, return_type:str="pandas", cache=True, results=[]):
         """
             :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param list placekeys:          Unique Placekey ID/IDs inside an array
@@ -414,9 +405,8 @@ class HTTP_Client:
         # save non weekly and monthly pattern first then the rest
         first_run = 1 # for the first pull, pull all data the rest only weekly
         data_frame = []
-        # print(f"\n\n\n\tlookup: {product=},{columns=},{date=},{return_type=}\n\n\n")
+        count = 0
         for i in self.date:
-            # print("\n\t "+i+"\n")
             if first_run:
                 dataset, data_type = self.__dataset(product, columns)
                 first_run = 0 
@@ -430,9 +420,8 @@ class HTTP_Client:
                 dataset = dataset.replace("_DATE_BLOCK_", "")
             else:
                 dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
-            # print(dataset+"\n")
-            __query__ = f"""query($placekeys: [Placekey!]) {{
- batch_lookup(placekeys: $placekeys) {{
+            __query__ = f"""query {{
+ batch_lookup(placekeys: {str(placekeys).replace("'", '"')}) {{
     placekey
     {dataset}
     }}
@@ -440,19 +429,43 @@ class HTTP_Client:
             if preview_query:
                 self.__prettyfy(__query__)
                 return
-            query = gql(__query__)            
-            result = self.client.execute(query, variable_values=params)
+            if cache:
+                query = gql(__query__)
+                results.append(query)
+                continue
+            #################################################################
+            # async with gql_Client(
+            #     transport=self.transport, fetch_schema_from_transport=True,
+            # ) as session:
+            #     result = await session.execute(query, variable_values=params)
+            #################################################################
+            try:
+                result = results[count] # result = await asyncio.wait_for(self.client.execute_async(query), timeout=30) # variable_values=params
+                count+=1
+            except TransportServerError:
+                print("*** TransportServerError => self.date:", self.date)
+                self.__adjustments(data_frame)
+                if self.return_type == "pandas":
+                    return self.df
+                elif self.return_type == "list":
+                    return self.lst
+                else:
+                    raise safeGraphError(f'return_type "{return_type}" does not exist')
+
             for place in result['batch_lookup']:
                 dict_ = {}
                 for j in data_type:
+                    print(place)
                     try:
                         dict_.update(place[j])
                         dict_['placekey'] = (place['placekey'])
-                    except TypeError:
+                    except (TypeError, KeyError):
+                        import pdb;pdb.set_trace()
                         # 'safegraph_weekly_patterns': None
                         pass
                 data_frame.append(dict_)
-
+        if cache:
+            return results
         self.__adjustments(data_frame)
         if self.return_type == "pandas":
             return self.df
@@ -461,17 +474,31 @@ class HTTP_Client:
         else:
             raise safeGraphError(f'return_type "{return_type}" does not exist')
 
-    def lookup_by_name(self, product:str, columns,
+    def lookup_by_name(self,product,columns,location_name=None,street_address=None,city=None,region=None,iso_country_code=None,postal_code=None,latitude=None,longitude=None,date='__default__',preview_query:bool=False,return_type='pandas'):
+        loop = asyncio.get_event_loop()
+        # asyncio.set_event_loop(loop)
+        results = loop.run_until_complete(self.lookup_by_name_(product,columns,location_name,street_address,city,region,iso_country_code,postal_code,latitude,longitude,date,preview_query,return_type, cache=True, results=[]))
+        if preview_query:
+            return
+        arr = [self.main_runner(query) for query in results]
+        results = asyncio.gather(*arr, return_exceptions=True)
+        report = loop.run_until_complete(results)
+        ret = loop.run_until_complete(self.lookup_by_name_(product,columns,location_name,street_address,city,region,iso_country_code,postal_code,latitude,longitude,date,preview_query,return_type, cache=False, results=report))
+        # loop.close()
+        return ret
+
+    async def lookup_by_name_(self, product:str, columns:list,
             location_name:str=None, 
             street_address:str=None, 
             city:str=None, 
             region:str=None,
             iso_country_code:str=None, 
             postal_code:str=None,
-            latitude:float=None, longitude:float=None, 
+            latitude:float=None, 
+            longitude:float=None, 
             date='__default__',
-            preview_query:str=False,
-            return_type:str="pandas"):
+            preview_query:bool=False,
+            return_type:str="pandas", cache=True, results=[]):
         """
             :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param columns:                 list or str 
@@ -521,9 +548,8 @@ When querying by location & address, it's necessary to have at least the followi
 """
         first_run = 1
         data_frame = []
-        # print(f"\n\n\n\tlookup_by_name: {product=},{columns=},{date=},{return_type=}\n\n\n")
+        count = 0
         for i in self.date:
-            # print("\n\t"+i+"\n")
             if first_run:
                 dataset, data_type = self.__dataset(product, columns)
                 first_run = 0 
@@ -537,7 +563,6 @@ When querying by location & address, it's necessary to have at least the followi
                 dataset = dataset.replace("_DATE_BLOCK_", "")
             else:
                 dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
-            # print(dataset+"\n")
             __query__ = f"""query {{
  lookup(query: {{{params}
  }}) {{ 
@@ -548,13 +573,32 @@ When querying by location & address, it's necessary to have at least the followi
             if preview_query:
                 self.__prettyfy(__query__)  
                 return
-            query = gql(__query__)
-            result = self.client.execute(query)
+            if cache:
+                query = gql(__query__)
+                results.append(query)
+                continue
+            #################################################################
+            # async with gql_Client(
+            #     transport=self.transport, fetch_schema_from_transport=True,
+            # ) as session:
+            #     result = await session.execute(query)
+            #################################################################
+            try:
+                result = results[count] # result = await asyncio.wait_for(self.client.execute_async(query), timeout=30)
+                count+=1
+            except TransportServerError:
+                self.__adjustments(data_frame)
+                if self.return_type == "pandas":
+                    return self.df
+                elif self.return_type == "list":
+                    return self.lst
+                else:
+                    raise safeGraphError(f'return_type "{return_type}" does not exist')
             dict_ = {}
             for j in data_type:
                 try:
                     dict_.update(result['lookup'][j])
-                except TypeError:
+                except (TypeError, KeyError):
                     # 'safegraph_weekly_patterns': None
                     pass
             try:
@@ -562,7 +606,8 @@ When querying by location & address, it's necessary to have at least the followi
             except TypeError:
                 raise safeGraphError("*** query didnt match with an item")
             data_frame.append(dict_)
-
+        if cache:
+            return results
         self.__adjustments(data_frame)
         if self.return_type == "pandas":
             return self.df
@@ -571,15 +616,26 @@ When querying by location & address, it's necessary to have at least the followi
         else:
             raise safeGraphError(f'return_type "{return_type}" does not exist')
 
-    def search(self, product, columns, date='__default__',
+    def search(self, product, columns, date='__default__',brand:str=None,brand_id:str=None,naics_code:int=None, phone_number:str=None,location_name:str=None, street_address:str=None, city:str=None, region:str=None, postal_code:str=None, iso_country_code:str=None,max_results:int=20,after_result_number:int=0,preview_query:bool=False,return_type:str="pandas"):
+        loop = asyncio.get_event_loop()
+        # asyncio.set_event_loop(loop)
+        results = loop.run_until_complete(self.search_(product,columns,date,brand,brand_id,naics_code,phone_number,location_name,street_address,city,region,postal_code,iso_country_code,max_results,after_result_number,preview_query,return_type, cache=True, results=[]))
+        if preview_query:
+            return
+        arr = [self.main_runner(query) for query in results]
+        results = asyncio.gather(*arr, return_exceptions=True)
+        report = loop.run_until_complete(results)
+        return loop.run_until_complete(self.search_(product,columns,date,brand,brand_id,naics_code,phone_number,location_name,street_address,city,region,postal_code,iso_country_code,max_results,after_result_number,preview_query,return_type, cache=False, results=report))
+
+    async def search_(self, product, columns, date='__default__',
         # params
         brand:str=None, brand_id:str=None, naics_code:int=None, phone_number:str=None,
         # address with following sub-fields
         location_name:str=None, street_address:str=None, city:str=None, region:str=None, postal_code:str=None, iso_country_code:str=None,
         max_results:int=20,
         after_result_number:int=0,
-        preview_query:str=False,
-        return_type:str="pandas"):
+        preview_query:bool=False,
+        return_type:str="pandas", cache=True, results=[]):
         """
             :param str product:             spesific product to pull data from either one core/geometry/weekly_patterns/monthly_patterns
             :param columns:                 list or str 
@@ -605,13 +661,11 @@ When querying by location & address, it's necessary to have at least the followi
             :return:                        The data of given placekey in return_type
             :rtype:                         pandas.DataFrame or dict
         """
-                                          ############
-        #################################################        |```|  /\   |`````|
-        self.max_results = max_results    ##################     |   | /  \  |     |
-        self.return_type = return_type    ###################    |`\` /____\ |     |
-        self.__date_setter(date)           #################      |  \/      \|     |
-        #################################################        |  /\       |\____|__
-                                          ############
+        ############
+        self.return_type = return_type
+        self.__date_setter(date)
+        ############
+        self.max_results = max_results
         product = f"safegraph_{product}.*"
         params = f"""
 {(lambda x,y: f' {x}: "{y}" ' if y!=None else "")("brand", brand)}
@@ -633,13 +687,15 @@ When querying by location & address, it's necessary to have at least the followi
         output = []
         chunks = self.__chunks()
         data_frame = []
-        # print(f"\n\n\n\tsearch: {columns=},{date=},{return_type=}\n\n\n")
+        arr = []
+        _max_after = False
+        count = 0
         for chu in chunks:
+            if _max_after:
+                break
             first = len(chu)
             first_run = 1 # for the first pull, pull all data the rest only weekly
-            data_frame = []
             for i in self.date:
-                # print("\n\t "+i+"\n")
                 if first_run:
                     dataset, data_type = self.__dataset(product, columns)
                     first_run = 0 
@@ -653,7 +709,6 @@ When querying by location & address, it's necessary to have at least the followi
                     dataset = dataset.replace("_DATE_BLOCK_", "")
                 else:
                     dataset = dataset.replace("_DATE_BLOCK_", f'(date: "{i}" )')
-                # print(dataset+"\n")
                 __query__ = f"""query {{
  search(first: {first} after: {after+after_result_number} filter: {{{params}   
  }}) {{ 
@@ -664,15 +719,54 @@ When querying by location & address, it's necessary to have at least the followi
                 if preview_query:
                     self.__prettyfy(__query__) 
                     return
-                query = gql(__query__)
+                if cache:
+                    query = gql(__query__)
+                    results.append(query)
+                    continue
                 try:
-                    result = self.client.execute(query)
-                except Exception as e:
-                    print("\n\n\n\t*** ERROR ***\n\n\n")
-                    raise e
-                    result = {'search': []}
+                    result = results[count]#await self.client.execute_async(query)
+                    count+=1
+                    result['search'][0]['placekey']
+                except IndexError:
                     self.errors.append(f"{after+after_result_number}-{first+after+after_result_number}")
-                output+=result['search']
+                    self.__lengthCheck__(data_frame)
+                    self.__adjustments(data_frame)
+                    self.__error_check(after_result_number)
+                    if self.return_type == "pandas":
+                        return self.df
+                    elif self.return_type == "list":
+                        return self.lst
+                    else:
+                        raise safeGraphError(f'return_type "{return_type}" does not exist')
+                except TransportServerError:
+                    print("*** TransportServerError")
+                    self.errors.append(f"{after+after_result_number}-{first+after+after_result_number}")
+                    self.__lengthCheck__(data_frame)
+                    self.__adjustments(data_frame)
+                    self.__error_check(after_result_number)
+                    print(self.return_type)
+                    if self.return_type == "pandas":
+                        return self.df
+                    elif self.return_type == "list":
+                        return self.lst
+                    else:
+                        raise safeGraphError(f'return_type "{return_type}" does not exist')
+                except Exception as e:
+                    try:
+                        if result.errors[0]['message'].startswith("Maximum 'after' value exceeded"):
+                            _max_after = True
+                            break
+                    except:
+                        pass
+                    for err in e.errors:
+                        if err["message"] == "Maximum 'after' value exceeded, use 'last_seen_placekey' to get next page. (max_offset=5000)":
+                            # {'message': "Maximum 'after' value exceeded, use 'last_seen_placekey' to get next page. (max_offset=5000)", 'path': ['search'], 'locations': [{'line': 2, 'column': 3}]}
+                            _max_after = True
+                            break
+                try:
+                    output+=result['search']
+                except UnboundLocalError:
+                    break
                 after += first
             for out in output:
                 dict_ = {}
@@ -680,9 +774,11 @@ When querying by location & address, it's necessary to have at least the followi
                     try:
                         dict_.update(out[j])
                         dict_['placekey'] = (out['placekey'])
-                    except TypeError:
+                    except (TypeError, KeyError):
                         pass
                 data_frame.append(dict_)
+        if cache:
+            return results
         self.__lengthCheck__(data_frame)
         self.__adjustments(data_frame)
         self.__error_check(after_result_number)
@@ -720,3 +816,10 @@ When querying by location & address, it's necessary to have at least the followi
             }}"""
         )
 
+    def get_patterns_supplemental(self, file:str, # home_panel_summary, normalization_stats, and visit_panel_summary
+        region:str, iso_country_code:str, date="__default__", 
+        max_hps_results:str='all', # This is only relevant when file = 'home_panel_summary'
+        after_hps_result:int=0, # This is only relevant when file = 'home_panel_summary'
+        preview_query:bool=False,):
+        pass
+
